classdef HMSGLargeClassReverberator < matlab.System
%PLATECLASSREVERBERATOR Griesinger's plate-class reverberation structure as
%described by Dattorro in [AES Journal Dec 97]
%Note that the defaults used here differ from those in the original paper,
%because they should match the defaults in audio.reverberator.

% Copyright 2015-2020 The MathWorks, Inc.

%#codegen
    %----------------------------------------------------------------------
    % PUBLIC TUNABLE PROPERTIES
    %----------------------------------------------------------------------
    properties
        RT60 = 0.1;    
        SampleRate = 44100;  
    end
    
    %----------------------------------------------------------------------
    % PUBLIC NONTUNABLE PROPERTIES
    %----------------------------------------------------------------------
    properties(Nontunable)        
        MaximumSampleRate = 192000; 
    end
    
    %----------------------------------------------------------------------
    % PRIVATE NONTUNABLE PROPERTIES
    %----------------------------------------------------------------------
    properties (Access = private)
        pLSumDelayTime
        pRSumDelayTime
        LInput_gain
        RInput_gain
        pStates
        pSRIndependentProperties
        pSampleRateCache
        pRT60Cache
    end
    
    %----------------------------------------------------------------------
    % PRIVATE NONTUNABLE PROPERTIES
    %----------------------------------------------------------------------
    properties (Access = private, Nontunable)      
        % Allpasses        
        pAP1DelayTime
        pAP2DelayTime

        pLAP1DelayLength
        pLAP2DelayLength
        pRAP1DelayLength
        pRAP2DelayLength

        A1_gain = 0.3
        A2_gain = 0.3

        % Delay
        pD1DelayTime
        pD2DelayTime
        pD3DelayTime
        pD4DelayTime

        pLD1DelayLength
        pLD2DelayLength
        pLD3DelayLength
        pLD4DelayLength
        pRD1DelayLength
        pRD2DelayLength
        pRD3DelayLength
        pRD4DelayLength


        % Single Nested Allpasses
        pSNOAPDelayTime
        pSNIAPDelayTime

        pLSNOAPDelayLength
        pLSNIAPDelayLength
        pRSNOAPDelayLength
        pRSNIAPDelayLength
   
        SNOA_gain = 0.5
        SNIA_gain = 0.25
        
        % Double Nested Allpasses
        pDNOAPDelayTime
        pDNIAP1DelayTime
        pDNIAP2DelayTime

        pLDNOAPDelayLength
        pLDNIAP1DelayLength
        pLDNIAP2DelayLength
        pRDNOAPDelayLength
        pRDNIAP1DelayLength
        pRDNIAP2DelayLength

        DNOA_gain = 0.5
        DNIA1_gain = 0.25
        DNIA2_gain = 0.25

        % Lowpasses
        LP1_gain = 0.6
        LP2_gain = 1/2600              

        pLSumDelayLength
        pRSumDelayLength
        temp = 15
        Out_gain = 0.34

        pDelvar = 0.1
        
    end
    
    %----------------------------------------------------------------------
    % PROTECTED METHODS
    %----------------------------------------------------------------------
    methods (Access = protected)
        function setupImpl(obj)            
            obj.pAP1DelayTime = 8;
            obj.pAP2DelayTime = 12;

            obj.pD1DelayTime = 4;
            obj.pD2DelayTime = 17;
            obj.pD3DelayTime = 31;
            obj.pD4DelayTime = 3;

            obj.pSNOAPDelayTime = 87;
            obj.pSNIAPDelayTime = 62;

            obj.pDNOAPDelayTime = 120;
            obj.pDNIAP1DelayTime = 76;
            obj.pDNIAP2DelayTime = 30;

            % Compute values once (these do not depend on tunable properties)
            obj.pLAP1DelayLength = round(obj.pAP1DelayTime * obj.temp);
            obj.pLAP2DelayLength = round(obj.pAP2DelayTime * obj.temp);

            obj.pLD1DelayLength = round(obj.pD1DelayTime * obj.temp);
            obj.pLD2DelayLength = round(obj.pD2DelayTime * obj.temp);
            obj.pLD3DelayLength = round(obj.pD3DelayTime * obj.temp);
            obj.pLD4DelayLength = round(obj.pD4DelayTime * obj.temp);

            obj.pLSNOAPDelayLength = round(obj.pSNOAPDelayTime * obj.temp);
            obj.pLSNIAPDelayLength = round(obj.pSNIAPDelayTime * obj.temp);

            obj.pLDNOAPDelayLength = round(obj.pDNOAPDelayTime * obj.temp);
            obj.pLDNIAP1DelayLength = round(obj.pDNIAP1DelayTime * obj.temp);
            obj.pLDNIAP2DelayLength = round(obj.pDNIAP2DelayTime * obj.temp);

            obj.pLSumDelayLength = obj.pLAP1DelayLength + obj.pLAP2DelayLength + obj.pLD1DelayLength + obj.pLD2DelayLength + obj.pLSNOAPDelayLength + obj.pLSNIAPDelayLength + obj.pLD3DelayLength + obj.pLD4DelayLength + obj.pLDNOAPDelayLength + obj.pLDNIAP1DelayLength + obj.pLDNIAP2DelayLength;
            obj.pLSumDelayTime = obj.pLSumDelayLength/obj.MaximumSampleRate;
            obj.LInput_gain = 0.001^(obj.pLSumDelayTime/(0.5*obj.RT60));

            obj.pRAP1DelayLength = round(obj.pAP1DelayTime * obj.temp * (1 + obj.pDelvar));
            obj.pRAP2DelayLength = round(obj.pAP2DelayTime * obj.temp * (1 + obj.pDelvar));

            obj.pRD1DelayLength = round(obj.pD1DelayTime * obj.temp * (1 + obj.pDelvar));
            obj.pRD2DelayLength = round(obj.pD2DelayTime * obj.temp * (1 + obj.pDelvar));
            obj.pRD3DelayLength = round(obj.pD3DelayTime * obj.temp * (1 + obj.pDelvar));
            obj.pRD4DelayLength = round(obj.pD4DelayTime * obj.temp * (1 + obj.pDelvar));

            obj.pRSNOAPDelayLength = round(obj.pSNOAPDelayTime * obj.temp * (1 + obj.pDelvar));
            obj.pRSNIAPDelayLength = round(obj.pSNIAPDelayTime * obj.temp * (1 + obj.pDelvar));

            obj.pRDNOAPDelayLength = round(obj.pDNOAPDelayTime * obj.temp * (1 + obj.pDelvar));
            obj.pRDNIAP1DelayLength = round(obj.pDNIAP1DelayTime * obj.temp * (1 + obj.pDelvar));
            obj.pRDNIAP2DelayLength = round(obj.pDNIAP2DelayTime * obj.temp * (1 + obj.pDelvar));

            obj.pRSumDelayLength = (obj.pRAP1DelayLength + obj.pRAP2DelayLength + obj.pRD1DelayLength + obj.pRD2DelayLength + obj.pRSNOAPDelayLength + obj.pRSNIAPDelayLength + obj.pRD3DelayLength + obj.pRD4DelayLength + obj.pRDNOAPDelayLength + obj.pRDNIAP1DelayLength + obj.pRDNIAP2DelayLength) * (1 + obj.pDelvar);
            obj.pRSumDelayTime = obj.pRSumDelayLength/obj.MaximumSampleRate;
            obj.RInput_gain = 0.001^(obj.pRSumDelayTime/(0.5*obj.RT60));
     
            resetPropertiesAndStates(obj);
        end
        %------------------------------------------------------------------
        function resetImpl(obj)
            resetPropertiesAndStates(obj);
        end
        %------------------------------------------------------------------
        function resetPropertiesAndStates(obj)           
            % Cache sample rate
            SR = obj.SampleRate;
            obj.pSampleRateCache = SR;
            ReverbT60 = obj.RT60;
            obj.pRT60Cache = ReverbT60;
            
            % Reset properties and states structures
            setSRIndependentProperties(obj);
            initializeStates(obj, SR);
        end
        %------------------------------------------------------------------
        function y = stepImpl(obj,u)
            if isempty(coder.target)
                % Call MEX implementation
                [y, obj.pStates] = applyHMSGClassReverbLarge_mex(u, ...
                    obj.pSRIndependentProperties, obj.pStates);
            else
                % Call MATLAB implementation
                [y, obj.pStates] = applyHMSGClassReverbLarge(u, ...
                    obj.pSRIndependentProperties, obj.pStates);            
            end
        end
        %------------------------------------------------------------------
        function processTunedPropertiesImpl(obj)
            % Process tuned properties method for pcr object
            
            % Local copies
            newRT60 = obj.RT60;
            newSR = obj.SampleRate;
            states = obj.pStates;
            
            % Save public parameters in structure
            setSRIndependentProperties(obj);

            if obj.pSampleRateCache ~= newSR || obj.pRT60Cache ~= newRT60
                % Need to update states and also the properties that depend
                % on sample rate          

                obj.pLSumDelayTime = obj.pLSumDelayLength/newSR;
                obj.pSRIndependentProperties.LInput_gain = 0.001^(obj.pLSumDelayTime/(0.5*newRT60));  

                obj.pRSumDelayTime = obj.pRSumDelayLength/newSR;
                obj.pSRIndependentProperties.RInput_gain = 0.001^(obj.pRSumDelayTime/(0.5*newRT60));  

                obj.pSampleRateCache = newSR;
                obj.pRT60Cache = newRT60;
            end

            % Save new states
            obj.pStates = states;
        end
        %------------------------------------------------------------------
        function s = saveObjectImpl(obj)
            % Default implementation saves all public properties
            s = saveObjectImpl@matlab.System(obj);
            if isLocked(obj)
                s.pSampleRateCache         = obj.pSampleRateCache;
                s.pRT60Cache               = obj.pRT60Cache;
                s.pSRIndependentProperties = obj.pSRIndependentProperties;
                s.pStates                  = obj.pStates;
                
                s.pAP1DelayTime = obj.pAP1DelayTime;
                s.pAP2DelayTime = obj.pAP2DelayTime;

                s.pD1DelayTime = obj.pD1DelayTime;
                s.pD2DelayTime = obj.pD2DelayTime;
                s.pD3DelayTime = obj.pD3DelayTime;
                s.pD4DelayTime = obj.pD4DelayTime;

                s.pSNOAPDelayTime = obj.pSNOAPDelayTime;
                s.pSNIAPDelayTime = obj.pSNIAPDelayTime;

                s.pDNOAPDelayTime = obj.pDNOAPDelayTime;
                s.pDNIAP1DelayTime = obj.pDNIAP1DelayTime;
                s.pDNIAP2DelayTime = obj.pDNIAP2DelayTime;

                s.pLSumDelayTime = obj.pLSumDelayTime;
                s.pRSumDelayTime = obj.pRSumDelayTime;

                s.pLAP1DelayLength = obj.pLAP1DelayLength;
                s.pLAP2DelayLength = obj.pLAP2DelayLength;

                s.pLD1DelayLength = obj.pLD1DelayLength;
                s.pLD2DelayLength = obj.pLD2DelayLength;
                s.pLD3DelayLength = obj.pLD3DelayLength;
                s.pLD4DelayLength = obj.pLD4DelayLength;

                s.pLSNOAPDelayLength = obj.pLSNOAPDelayLength;
                s.pLSNIAPDelayLength = obj.pLSNIAPDelayLength;
                
                s.pLDNOAPDelayLength = obj.pLDNOAPDelayLength;
                s.pLDNIAP1DelayLength = obj.pLDNIAP1DelayLength;
                s.pLDNIAP2DelayLength = obj.pLDNIAP2DelayLength;

                s.pLSumDelayLength = obj.pLSumDelayLength;

                s.pRD1DelayLength = obj.pRD1DelayLength;
                s.pRD2DelayLength = obj.pRD2DelayLength;
                s.pRD3DelayLength = obj.pRD3DelayLength;
                s.pRD4DelayLength = obj.pRD4DelayLength;

                s.pRSNOAPDelayLength = obj.pRSNOAPDelayLength;
                s.pRSNIAPDelayLength = obj.pRSNIAPDelayLength;
                
                s.pRDNOAPDelayLength = obj.pRDNOAPDelayLength;
                s.pRDNIAP1DelayLength = obj.pRDNIAP1DelayLength;
                s.pRDNIAP2DelayLength = obj.pRDNIAP2DelayLength;

                s.pRSumDelayLength = obj.pRSumDelayLength;
            end
        end
        %------------------------------------------------------------------
        function loadObjectImpl(obj, s, wasLocked)
            % Re-load state if saved version was locked
            if wasLocked
                obj.pSampleRateCache         = s.pSampleRateCache;
                obj.pRT60Cache               = s.pRT60Cache;
                obj.pSRIndependentProperties   = s.pSRIndependentProperties;
                obj.pStates                  = s.pStates;

                obj.pAP1DelayTime = s.pAP1DelayTime;
                obj.pAP2DelayTime = s.pAP2DelayTime;

                obj.pD1DelayTime = s.pD1DelayTime;
                obj.pD2DelayTime = s.pD2DelayTime;
                obj.pD3DelayTime = s.pD3DelayTime;
                obj.pD4DelayTime = s.pD4DelayTime;
                
                obj.pSNOAPDelayTime = s.pSNOAPDelayTime;
                obj.pSNIAPDelayTime = s.pSNIAPDelayTime;

                obj.pDNOAPDelayTime = s.pDNOAPDelayTime;
                obj.pDNIAP1DelayTime = s.pDNIAP1DelayTime;
                obj.pDNIAP2DelayTime = s.pDNIAP2DelayTime; 

                obj.pLSumDelayTime = s.pLSumDelayTime;

                obj.pRSumDelayTime = s.pRSumDelayTime;

                obj.pLAP1DelayLength = s.pAP1DelayLength;
                obj.pLAP2DelayLength = s.pAP2DelayLength;

                obj.pLD1DelayLength = s.pLD1DelayLength;
                obj.pLD2DelayLength = s.pLD2DelayLength;
                obj.pLD3DelayLength = s.pLD3DelayLength;
                obj.pLD4DelayLength = s.pLD4DelayLength;

                obj.pLSNOAPDelayLength = s.pLSNOAPDelayLength;
                obj.pLSNIAPDelayLength = s.pLSNIAPDelayLength;

                obj.pLDNOAPDelayLength = s.pLDNOAPDelayLength;
                obj.pLDNIAP1DelayLength = s.pLDNIAP1DelayLength;
                obj.pLDNIAP2DelayLength = s.pLDNIAP2DelayLength;

                obj.pLSumDelayLength = s.pLSumDelayLength;

                obj.pRAP1DelayLength = s.pAP1DelayLength;
                obj.pRAP2DelayLength = s.pAP2DelayLength;

                obj.pRD1DelayLength = s.pRD1DelayLength;
                obj.pRD2DelayLength = s.pRD2DelayLength;
                obj.pRD3DelayLength = s.pRD3DelayLength;
                obj.pRD4DelayLength = s.pRD4DelayLength;

                obj.pRSNOAPDelayLength = s.pRSNOAPDelayLength;
                obj.pRSNIAPDelayLength = s.pRSNIAPDelayLength;

                obj.pRDNOAPDelayLength = s.pRDNOAPDelayLength;
                obj.pRDNIAP1DelayLength = s.pRDNIAP1DelayLength;
                obj.pRDNIAP2DelayLength = s.pRDNIAP2DelayLength;

                obj.pRSumDelayLength = s.pRSumDelayLength;

            end
            % Call base class method
            loadObjectImpl@matlab.System(obj,s,wasLocked);
        end
        %------------------------------------------------------------------
       
        %------------------------------------------------------------------
        function setSRIndependentProperties(obj)
            % Properties that depend on sample rate
            
            obj.pSRIndependentProperties = struct(...
                'LInput_gain', 0.001^(obj.pLSumDelayTime/(0.5*obj.RT60)),...
                'RInput_gain', 0.001^(obj.pRSumDelayTime/(0.5*obj.RT60)),...
                'A1_gain', 0.3,...
                'A2_gain', 0.3,...
                'SNOA_gain', 0.5,...
                'SNIA_gain', 0.25,...
                'DNOA_gain', 0.5,...
                'DNIA1_gain', 0.25,...
                'DNIA2_gain', 0.25,...
                'LP1_gain', 0.6,...
                'LP2_gain', 1/2600,...
                'Out_gain', 0.34...
                );
        end

        %------------------------------------------------------------------
        function initializeStates(obj, SR)
            obj.pLSumDelayTime = obj.pLSumDelayLength/SR;
            obj.LInput_gain = 0.001^(obj.pLSumDelayTime/(0.5*obj.RT60));

            obj.pRSumDelayTime = obj.pRSumDelayLength/SR;
            obj.RInput_gain = 0.001^(obj.pRSumDelayTime/(0.5*obj.RT60));

            % Initialize or reset states
            obj.A1_gain = 0.3;
            obj.A2_gain = 0.3;
            obj.SNOA_gain = 0.5;
            obj.SNIA_gain = 0.25;
            obj.DNOA_gain = 0.5;
            obj.DNIA1_gain = 0.25;
            obj.DNIA2_gain = 0.25;
            obj.LP1_gain = 0.6;
            obj.LP2_gain = 1/2600;
            obj.Out_gain = 0.34;
            
            obj.pStates = struct(...      
                'LAP1DelayLine', zeros(obj.pLAP1DelayLength, 1), ...
                'LAP1ReadPtr',   1, ...
                'LAP1WritePtr',  checkUpperBound(obj,obj.pLAP1DelayLength+1,obj.pLAP1DelayLength), ...
                'LAP2DelayLine', zeros(obj.pLAP2DelayLength, 1), ...
                'LAP2ReadPtr',   1, ...
                'LAP2WritePtr',  checkUpperBound(obj,obj.pLAP2DelayLength+1,obj.pLAP2DelayLength), ...
                'LD1DelayLine',  zeros(obj.pLD1DelayLength, 1), ...
                'LD1ReadPtr',    1, ...
                'LD1WritePtr',   checkUpperBound(obj,obj.pLD1DelayLength+1,obj.pLD1DelayLength), ...
                'LD2DelayLine',  zeros(obj.pLD2DelayLength, 1), ...
                'LD2ReadPtr',    1, ...
                'LD2WritePtr',   checkUpperBound(obj,obj.pLD2DelayLength+1,obj.pLD2DelayLength), ...
                'LD3DelayLine',  zeros(obj.pLD3DelayLength, 1), ...
                'LD3ReadPtr',    1, ...
                'LD3WritePtr',   checkUpperBound(obj,obj.pLD3DelayLength+1,obj.pLD3DelayLength), ...
                'LD4DelayLine',  zeros(obj.pLD4DelayLength, 1), ...
                'LD4ReadPtr',    1, ...
                'LD4WritePtr',   checkUpperBound(obj,obj.pLD4DelayLength+1,obj.pLD4DelayLength), ...
                'LSNOAPDelayLine', zeros(obj.pLSNOAPDelayLength, 1), ...
                'LSNOAPReadPtr',   1, ...
                'LSNOAPWritePtr',  checkUpperBound(obj,obj.pLSNOAPDelayLength+1,obj.pLSNOAPDelayLength), ...
                'LSNIAPDelayLine', zeros(obj.pLSNIAPDelayLength, 1), ...
                'LSNIAPReadPtr',   1, ...
                'LSNIAPWritePtr',  checkUpperBound(obj,obj.pLSNIAPDelayLength+1,obj.pLSNIAPDelayLength), ...
                'LDNOAPDelayLine', zeros(obj.pLDNOAPDelayLength, 1), ...
                'LDNOAPReadPtr',   1, ...
                'LDNOAPWritePtr',  checkUpperBound(obj,obj.pLDNOAPDelayLength+1,obj.pLDNOAPDelayLength), ...
                'LDNIAP1DelayLine', zeros(obj.pLDNIAP1DelayLength, 1), ...
                'LDNIAP1ReadPtr',   1, ...
                'LDNIAP1WritePtr',  checkUpperBound(obj,obj.pLDNIAP1DelayLength+1,obj.pLDNIAP1DelayLength), ...
                'LDNIAP2DelayLine', zeros(obj.pLDNIAP2DelayLength, 1), ...
                'LDNIAP2ReadPtr',   1, ...
                'LDNIAP2WritePtr',  checkUpperBound(obj,obj.pLDNIAP2DelayLength+1,obj.pLDNIAP2DelayLength),...              
                'LLP1DelayLine', 0, ...
                'LLP2DelayLine', 0, ...
                'RAP1DelayLine', zeros(obj.pRAP1DelayLength, 1), ...
                'RAP1ReadPtr',   1, ...
                'RAP1WritePtr',  checkUpperBound(obj,obj.pRAP1DelayLength+1,obj.pRAP1DelayLength), ...
                'RAP2DelayLine', zeros(obj.pRAP2DelayLength, 1), ...
                'RAP2ReadPtr',   1, ...
                'RAP2WritePtr',  checkUpperBound(obj,obj.pRAP2DelayLength+1,obj.pRAP2DelayLength), ...
                'RD1DelayLine',  zeros(obj.pRD1DelayLength, 1), ...
                'RD1ReadPtr',    1, ...
                'RD1WritePtr',   checkUpperBound(obj,obj.pRD1DelayLength+1,obj.pRD1DelayLength), ...
                'RD2DelayLine',  zeros(obj.pRD2DelayLength, 1), ...
                'RD2ReadPtr',    1, ...
                'RD2WritePtr',   checkUpperBound(obj,obj.pRD2DelayLength+1,obj.pRD2DelayLength), ...
                'RD3DelayLine',  zeros(obj.pRD3DelayLength, 1), ...
                'RD3ReadPtr',    1, ...
                'RD3WritePtr',   checkUpperBound(obj,obj.pRD3DelayLength+1,obj.pRD3DelayLength), ...
                'RD4DelayLine',  zeros(obj.pRD4DelayLength, 1), ...
                'RD4ReadPtr',    1, ...
                'RD4WritePtr',   checkUpperBound(obj,obj.pRD4DelayLength+1,obj.pRD4DelayLength), ...
                'RSNOAPDelayLine', zeros(obj.pRSNOAPDelayLength, 1), ...
                'RSNOAPReadPtr',   1, ...
                'RSNOAPWritePtr',  checkUpperBound(obj,obj.pRSNOAPDelayLength+1,obj.pRSNOAPDelayLength), ...
                'RSNIAPDelayLine', zeros(obj.pRSNIAPDelayLength, 1), ...
                'RSNIAPReadPtr',   1, ...
                'RSNIAPWritePtr',  checkUpperBound(obj,obj.pRSNIAPDelayLength+1,obj.pRSNIAPDelayLength), ...
                'RDNOAPDelayLine', zeros(obj.pRDNOAPDelayLength, 1), ...
                'RDNOAPReadPtr',   1, ...
                'RDNOAPWritePtr',  checkUpperBound(obj,obj.pRDNOAPDelayLength+1,obj.pRDNOAPDelayLength), ...
                'RDNIAP1DelayLine', zeros(obj.pRDNIAP1DelayLength, 1), ...
                'RDNIAP1ReadPtr',   1, ...
                'RDNIAP1WritePtr',  checkUpperBound(obj,obj.pRDNIAP1DelayLength+1,obj.pRDNIAP1DelayLength), ...
                'RDNIAP2DelayLine', zeros(obj.pRDNIAP2DelayLength, 1), ...
                'RDNIAP2ReadPtr',   1, ...
                'RDNIAP2WritePtr',  checkUpperBound(obj,obj.pRDNIAP2DelayLength+1,obj.pRDNIAP2DelayLength),...
                'RLP1DelayLine', 0, ...
                'RLP2DelayLine', 0 ...
                );
        end
        %------------------------------------------------------------------
        function ptr = checkUpperBound(~, ptr, maxVal)
            % Wrap overflow of ptr around [1...maxval]
            if ptr > maxVal
                ptr = ptr - maxVal;
            end
        end
        %------------------------------------------------------------------
        function ptr = checkLowerBound(~, ptr, maxVal)
            % Wrap underflow of ptr around [1...maxval]
            if ptr < 1
                ptr = ptr + maxVal;
            end
        end
        %------------------------------------------------------------------
    end
end