classdef HMSGSmallClassReverberator < matlab.System
%PLATECLASSREVERBERATOR Griesinger's plate-class reverberation structure as
%described by Dattorro in [AES Journal Dec 97]
%Note that the defaults used here differ from those in the original paper,
%because they should match the defaults in audio.reverberator.

% Copyright 2015-2020 The MathWorks, Inc.

%#codegen
    %----------------------------------------------------------------------
    % PUBLIC TUNABLE PROPERTIES
    %----------------------------------------------------------------------
    properties
        RT60 = 0.1;    
        SampleRate = 44100;  
    end
    
    %----------------------------------------------------------------------
    % PUBLIC NONTUNABLE PROPERTIES
    %----------------------------------------------------------------------
    properties(Nontunable)        
        MaximumSampleRate = 192000; 
    end
    
    %----------------------------------------------------------------------
    % PRIVATE NONTUNABLE PROPERTIES
    %----------------------------------------------------------------------
    properties (Access = private)
        C1_gain
        C2_gain
        C3_gain
        C4_gain

        pStates
        pSRIndependentProperties
        pSampleRateCache
        pRT60Cache
    end
    
    %----------------------------------------------------------------------
    % PRIVATE NONTUNABLE PROPERTIES
    %----------------------------------------------------------------------
    properties (Access = private, Nontunable)
        pC1DelayTime
        pC2DelayTime
        pC3DelayTime
        pC4DelayTime
       
        pAP1DelayTime
        pAP2DelayTime
        
        pDelvar = 0.1
        
        pMaxLC1DelayLength
        pMaxLC2DelayLength
        pMaxLC3DelayLength
        pMaxLC4DelayLength
        pMaxRC1DelayLength
        pMaxRC2DelayLength
        pMaxRC3DelayLength
        pMaxRC4DelayLength
        pMaxLAP1DelayLength
        pMaxLAP2DelayLength
        pMaxRAP1DelayLength
        pMaxRAP2DelayLength      

        A1_RT60 = 0.09683
        A2_RT60 = 0.03292

        A1_gain
        A2_gain

        LP_gain = 0.3;
        
    end
    
    %----------------------------------------------------------------------
    % PROTECTED METHODS
    %----------------------------------------------------------------------
    methods (Access = protected)
        function setupImpl(obj)
            % Cast to input type
            maxSR = obj.MaximumSampleRate;
            
            obj.pC1DelayTime = 0.0297;
            obj.pC2DelayTime = 0.0371;
            obj.pC3DelayTime = 0.0411;
            obj.pC4DelayTime = 0.0437;
            obj.pAP1DelayTime = 0.005;
            obj.pAP2DelayTime = 0.0017;

            % Compute values once (these do not depend on tunable properties)
            obj.pMaxLC1DelayLength = round(obj.pC1DelayTime * maxSR);
            obj.pMaxLC2DelayLength = round(obj.pC2DelayTime * maxSR);
            obj.pMaxLC3DelayLength = round(obj.pC3DelayTime * maxSR);
            obj.pMaxLC4DelayLength = round(obj.pC4DelayTime * maxSR);
            obj.pMaxLAP1DelayLength = round(obj.pAP1DelayTime * maxSR);
            obj.pMaxLAP2DelayLength = round(obj.pAP2DelayTime * maxSR);

            obj.pMaxRC1DelayLength = round(obj.pC1DelayTime * maxSR + obj.pDelvar * 297);
            obj.pMaxRC2DelayLength = round(obj.pC2DelayTime * maxSR + obj.pDelvar * 371);
            obj.pMaxRC3DelayLength = round(obj.pC3DelayTime * maxSR + obj.pDelvar * 411);
            obj.pMaxRC4DelayLength = round(obj.pC4DelayTime * maxSR + obj.pDelvar * 437);
            obj.pMaxRAP1DelayLength = round(obj.pAP1DelayTime * maxSR + obj.pDelvar * 50);
            obj.pMaxRAP2DelayLength = round(obj.pAP2DelayTime * maxSR + obj.pDelvar * 17);

            obj.C1_gain = 0.001^(obj.pC1DelayTime/obj.RT60);
            obj.C2_gain = 0.001^(obj.pC2DelayTime/obj.RT60);
            obj.C3_gain = 0.001^(obj.pC3DelayTime/obj.RT60);
            obj.C4_gain = 0.001^(obj.pC4DelayTime/obj.RT60); 

            obj.A1_gain = 0.001^(obj.pAP1DelayTime/obj.A1_RT60);
            obj.A2_gain = 0.001^(obj.pAP2DelayTime/obj.A2_RT60);
     
            resetPropertiesAndStates(obj);
        end
        %------------------------------------------------------------------
        function resetImpl(obj)
            resetPropertiesAndStates(obj);
        end
        %------------------------------------------------------------------
        function resetPropertiesAndStates(obj)          
            % Cache sample rate
            SR = obj.SampleRate;
            obj.pSampleRateCache = SR;
            ReverbT60 = obj.RT60;
            obj.pRT60Cache = ReverbT60;
            
            % Reset properties and states structures
            setSRIndependentProperties(obj);
            initializeStates(obj, SR);
        end
        %------------------------------------------------------------------
        function y = stepImpl(obj,u)
            if isempty(coder.target)
                % Call MEX implementation
                [y, obj.pStates] = applyHMSGClassReverbSmall_mex(u, ...
                    obj.pSRIndependentProperties, obj.pStates);
            else
                % Call MATLAB implementation
                [y, obj.pStates] = applyHMSGClassReverbSmall(u, ...
                    obj.pSRIndependentProperties, obj.pStates);            
            end
        end
        %------------------------------------------------------------------
        function processTunedPropertiesImpl(obj)
            % Process tuned properties method for pcr object
            
            % Local copies
            newRT60 = obj.RT60;
            newSR = obj.SampleRate;
            states = obj.pStates;
            
            % Save public parameters in structure
            setSRIndependentProperties(obj);

            if obj.pSampleRateCache ~= newSR
                % Need to update states and also the properties that depend
                % on sample rate
                LC1DelayLength = round(obj.pC1DelayTime*newSR);
                LC2DelayLength = round(obj.pC2DelayTime*newSR);
                LC3DelayLength = round(obj.pC3DelayTime*newSR);
                LC4DelayLength = round(obj.pC4DelayTime*newSR);
                RC1DelayLength = round(obj.pC1DelayTime*newSR + obj.pDelvar * 297);
                RC2DelayLength = round(obj.pC2DelayTime*newSR + obj.pDelvar * 371);
                RC3DelayLength = round(obj.pC3DelayTime*newSR + obj.pDelvar * 411);
                RC4DelayLength = round(obj.pC4DelayTime*newSR + obj.pDelvar * 437);
    
                LAP1DelayLength = round(obj.pAP1DelayTime*newSR);
                LAP2DelayLength = round(obj.pAP2DelayTime*newSR);
                RAP1DelayLength = round(obj.pAP1DelayTime*newSR + obj.pDelvar * 50);
                RAP2DelayLength = round(obj.pAP2DelayTime*newSR + obj.pDelvar * 17);

                % Delay line is already initialized to its max value. Just
                % need to reposition the read pointer according to new
                % delay length.
                states.LC1ReadPtr = checkLowerBound(obj, states.LC1WritePtr-LC1DelayLength, obj.pMaxLC1DelayLength);
                states.LC2ReadPtr = checkLowerBound(obj, states.LC2WritePtr-LC2DelayLength, obj.pMaxLC2DelayLength);
                states.LC3ReadPtr = checkLowerBound(obj, states.LC3WritePtr-LC3DelayLength, obj.pMaxLC3DelayLength);
                states.LC4ReadPtr = checkLowerBound(obj, states.LC4WritePtr-LC4DelayLength, obj.pMaxLC4DelayLength);
                states.RC1ReadPtr = checkLowerBound(obj, states.RC1WritePtr-RC1DelayLength, obj.pMaxRC1DelayLength);
                states.RC2ReadPtr = checkLowerBound(obj, states.RC2WritePtr-RC2DelayLength, obj.pMaxRC2DelayLength);
                states.RC3ReadPtr = checkLowerBound(obj, states.RC3WritePtr-RC3DelayLength, obj.pMaxRC3DelayLength);
                states.RC4ReadPtr = checkLowerBound(obj, states.RC4WritePtr-RC4DelayLength, obj.pMaxRC4DelayLength);

                states.LAP1ReadPtr = checkLowerBound(obj, states.LAP1WritePtr-LAP1DelayLength, obj.pMaxLAP1DelayLength);
                states.LAP2ReadPtr = checkLowerBound(obj, states.LAP2WritePtr-LAP2DelayLength, obj.pMaxLAP2DelayLength);
                states.RAP1ReadPtr = checkLowerBound(obj, states.RAP1WritePtr-RAP1DelayLength, obj.pMaxRAP1DelayLength);
                states.RAP2ReadPtr = checkLowerBound(obj, states.RAP2WritePtr-RAP2DelayLength, obj.pMaxRAP2DelayLength);

                obj.pSampleRateCache = newSR;
            end

            if obj.pRT60Cache ~= newRT60
                % Need to update states and also the properties that depend
                % on sample rate
                obj.pSRIndependentProperties.C1_gain = 0.001^(obj.pC1DelayTime/newRT60);
                obj.pSRIndependentProperties.C2_gain = 0.001^(obj.pC2DelayTime/newRT60);
                obj.pSRIndependentProperties.C3_gain = 0.001^(obj.pC3DelayTime/newRT60);
                obj.pSRIndependentProperties.C4_gain = 0.001^(obj.pC4DelayTime/newRT60);       
                
                obj.pRT60Cache = newRT60;
            end
            % Save new states
            obj.pStates = states;
        end
        %------------------------------------------------------------------
        function s = saveObjectImpl(obj)
            % Default implementation saves all public properties
            s = saveObjectImpl@matlab.System(obj);
            if isLocked(obj)
                s.pSampleRateCache         = obj.pSampleRateCache;
                s.pRT60Cache               = obj.pRT60Cache;
                s.pSRIndependentProperties = obj.pSRIndependentProperties;
                s.pStates                  = obj.pStates;
                
                s.pC1DelayTime = obj.pC1DelayTime;
                s.pC2DelayTime = obj.pC2DelayTime;
                s.pC3DelayTime = obj.pC3DelayTime;
                s.pC4DelayTime = obj.pC4DelayTime;
                s.pAP1DelayTime = obj.pAP1DelayTime;
                s.pAP2DelayTime = obj.pAP2DelayTime;
                
                s.pDelvar = obj.pDelvar;                              
                
                s.pMaxLC1DelayLength = obj.pMaxLC1DelayLength;
                s.pMaxLC2DelayLength = obj.pMaxLC2DelayLength;
                s.pMaxLC3DelayLength = obj.pMaxLC3DelayLength;
                s.pMaxLC4DelayLength = obj.pMaxLC4DelayLength;
                s.pMaxRC1DelayLength = obj.pMaxRC1DelayLength;
                s.pMaxRC2DelayLength = obj.pMaxRC2DelayLength;
                s.pMaxRC3DelayLength = obj.pMaxRC3DelayLength;
                s.pMaxRC4DelayLength = obj.pMaxRC4DelayLength;

                s.pMaxLAP1DelayLength = obj.pMaxLAP1DelayLength;
                s.pMaxLAP2DelayLength = obj.pMaxLAP2DelayLength;
                s.pMaxRAP1DelayLength = obj.pMaxRAP1DelayLength;
                s.pMaxRAP2DelayLength = obj.pMaxRAP2DelayLength;
     
            end
        end
        %------------------------------------------------------------------
        function loadObjectImpl(obj, s, wasLocked)
            % Re-load state if saved version was locked
            if wasLocked
                obj.pSampleRateCache         = s.pSampleRateCache;
                obj.pRT60Cache               = s.pRT60Cache;
                obj.pSRIndependentProperties   = s.pSRIndependentProperties;
                obj.pStates                  = s.pStates;
                
                obj.pC1DelayTime = s.pC1DelayTime;
                obj.pC2DelayTime = s.pC2DelayTime;
                obj.pC3DelayTime = s.pC3DelayTime;
                obj.pC4DelayTime = s.pC4DelayTime;
                obj.pAP1DelayTime = s.pAP1DelayTime;
                obj.pAP2DelayTime = s.pAP2DelayTime;             
                
                obj.pDelvar = s.pDelvar;              
                
                obj.pMaxLC1DelayLength = s.pMaxLC1DelayLength;
                obj.pMaxLC2DelayLength = s.pMaxLC2DelayLength;
                obj.pMaxLC3DelayLength = s.pMaxLC3DelayLength;
                obj.pMaxLC4DelayLength = s.pMaxLC4DelayLength;
                obj.pMaxRC1DelayLength = s.pMaxRC1DelayLength;
                obj.pMaxRC2DelayLength = s.pMaxRC2DelayLength;
                obj.pMaxRC3DelayLength = s.pMaxRC3DelayLength;
                obj.pMaxRC4DelayLength = s.pMaxRC4DelayLength;

                obj.pMaxLAP1DelayLength = s.pMaxLAP1DelayLength;
                obj.pMaxLAP2DelayLength = s.pMaxLAP2DelayLength;
                obj.pMaxRAP1DelayLength = s.pMaxRAP1DelayLength;
                obj.pMaxRAP2DelayLength = s.pMaxRAP2DelayLength;
            end
            % Call base class method
            loadObjectImpl@matlab.System(obj,s,wasLocked);
        end
        %------------------------------------------------------------------
       
        %------------------------------------------------------------------
        function setSRIndependentProperties(obj)
            % Properties that depend on sample rate
            
            obj.pSRIndependentProperties = struct(...
                'LP_gain', 0.3,...
                'C1_gain', 0.001^(obj.pC1DelayTime/obj.RT60),...
                'C2_gain', 0.001^(obj.pC2DelayTime/obj.RT60),...
                'C3_gain', 0.001^(obj.pC3DelayTime/obj.RT60),...
                'C4_gain', 0.001^(obj.pC4DelayTime/obj.RT60),...
                'A1_gain', 0.001^(obj.pAP1DelayTime/obj.A1_RT60),...
                'A2_gain', 0.001^(obj.pAP2DelayTime/obj.A2_RT60));
        end

        %------------------------------------------------------------------
        function initializeStates(obj, SR)
            % Initialize or reset states
            LC1DelayLength = round(obj.pC1DelayTime*SR);
            LC2DelayLength = round(obj.pC2DelayTime*SR);
            LC3DelayLength = round(obj.pC3DelayTime*SR);
            LC4DelayLength = round(obj.pC4DelayTime*SR);
            RC1DelayLength = round(obj.pC1DelayTime*SR+obj.pDelvar*297);
            RC2DelayLength = round(obj.pC2DelayTime*SR+obj.pDelvar*371);
            RC3DelayLength = round(obj.pC3DelayTime*SR+obj.pDelvar*411);
            RC4DelayLength = round(obj.pC4DelayTime*SR+obj.pDelvar*437);

            LAP1DelayLength = round(obj.pAP1DelayTime*SR);
            LAP2DelayLength = round(obj.pAP2DelayTime*SR);
            RAP1DelayLength = round(obj.pAP1DelayTime*SR+obj.pDelvar*50);
            RAP2DelayLength = round(obj.pAP2DelayTime*SR+obj.pDelvar*17);

            obj.C1_gain = 0.001^(obj.pC1DelayTime/obj.RT60);
            obj.C2_gain = 0.001^(obj.pC2DelayTime/obj.RT60);
            obj.C3_gain = 0.001^(obj.pC3DelayTime/obj.RT60);
            obj.C4_gain = 0.001^(obj.pC4DelayTime/obj.RT60); 

            obj.A1_gain = 0.001^(obj.pAP1DelayTime/obj.A1_RT60);
            obj.A2_gain = 0.001^(obj.pAP2DelayTime/obj.A2_RT60);

            obj.LP_gain = 0.3;
            
            obj.pStates = struct(...
                'LLP1DelayLine', 0, ...
                'RLP1DelayLine', 0, ...
                'LC1DelayLine',  zeros(obj.pMaxLC1DelayLength, 1), ...
                'LC1ReadPtr',    1, ...
                'LC1WritePtr',   checkUpperBound(obj,LC1DelayLength+1,obj.pMaxLC1DelayLength), ...
                'LC2DelayLine',  zeros(obj.pMaxLC2DelayLength, 1), ...
                'LC2ReadPtr',    1, ...
                'LC2WritePtr',   checkUpperBound(obj,LC2DelayLength+1,obj.pMaxLC2DelayLength), ...
                'LC3DelayLine',  zeros(obj.pMaxLC3DelayLength, 1), ...
                'LC3ReadPtr',    1, ...
                'LC3WritePtr',   checkUpperBound(obj,LC3DelayLength+1,obj.pMaxLC3DelayLength), ...
                'LC4DelayLine',  zeros(obj.pMaxLC4DelayLength, 1), ...
                'LC4ReadPtr',    1, ...
                'LC4WritePtr',   checkUpperBound(obj,LC4DelayLength+1,obj.pMaxLC4DelayLength), ...
                'RC1DelayLine',  zeros(obj.pMaxRC1DelayLength, 1), ...
                'RC1ReadPtr',    1, ...
                'RC1WritePtr',   checkUpperBound(obj,RC1DelayLength+1,obj.pMaxRC1DelayLength), ...
                'RC2DelayLine',  zeros(obj.pMaxRC2DelayLength, 1), ...
                'RC2ReadPtr',    1, ...
                'RC2WritePtr',   checkUpperBound(obj,RC2DelayLength+1,obj.pMaxRC2DelayLength), ...
                'RC3DelayLine',  zeros(obj.pMaxRC3DelayLength, 1), ...
                'RC3ReadPtr',    1, ...
                'RC3WritePtr',   checkUpperBound(obj,RC3DelayLength+1,obj.pMaxRC3DelayLength), ...
                'RC4DelayLine',  zeros(obj.pMaxRC4DelayLength, 1), ...
                'RC4ReadPtr',    1, ...
                'RC4WritePtr',   checkUpperBound(obj,RC4DelayLength+1,obj.pMaxRC4DelayLength), ...                        
                'LAP1DelayLine', zeros(obj.pMaxLAP1DelayLength, 1), ...
                'LAP1ReadPtr',   1, ...
                'LAP1WritePtr',  checkUpperBound(obj,LAP1DelayLength+1,obj.pMaxLAP1DelayLength), ...
                'LAP2DelayLine', zeros(obj.pMaxLAP2DelayLength, 1), ...
                'LAP2ReadPtr',   1, ...
                'LAP2WritePtr',  checkUpperBound(obj,LAP2DelayLength+1,obj.pMaxLAP2DelayLength), ...
                'RAP1DelayLine', zeros(obj.pMaxRAP1DelayLength, 1), ...
                'RAP1ReadPtr',   1, ...
                'RAP1WritePtr',  checkUpperBound(obj,RAP1DelayLength+1,obj.pMaxRAP1DelayLength), ...
                'RAP2DelayLine', zeros(obj.pMaxRAP2DelayLength, 1), ...
                'RAP2ReadPtr',   1, ...
                'RAP2WritePtr',  checkUpperBound(obj,RAP2DelayLength+1,obj.pMaxRAP2DelayLength)...            
                );
        end
        %------------------------------------------------------------------
        function ptr = checkUpperBound(~, ptr, maxVal)
            % Wrap overflow of ptr around [1...maxval]
            if ptr > maxVal
                ptr = ptr - maxVal;
            end
        end
        %------------------------------------------------------------------
        function ptr = checkLowerBound(~, ptr, maxVal)
            % Wrap underflow of ptr around [1...maxval]
            if ptr < 1
                ptr = ptr + maxVal;
            end
        end
        %------------------------------------------------------------------
    end
end